import requests
import copy
import os
import sqlite3
import time
from pathlib import Path
from tool.tool_config import get_cache_manager, make_github_request, clone_repo, get_last_page_info
import git
import logging

cache_manager = get_cache_manager()


def get_repo_author_commits(api_url):
    # Since we can't return the commits in ascending date order, we'll just return the latest commit
    # This response also holds the number of pages, so the last page will have the first commit
    search_url = f"{api_url}&per_page=1"
    last_page = get_last_page_info(search_url, max_retries=2, retry_delay=2, sleep_between_requests=2)
    if not last_page:
        return None

    last_page_url = f"{search_url}&page={last_page}"
    return make_github_request(last_page_url, max_retries=2, retry_delay=2, sleep_between_requests=2)


def get_user_first_commit_info(data):
    """
    Get the first commit information for each author in the given data.

    Args:
        data (dict): A dictionary containing package information with authors.

    Returns:
        dict: A dictionary with updated package information including first commit details.
    """
    cache_manager._setup_requests_cache("get_user_commit_info")
    logging.info("Getting user commit information")

    packages_data = copy.deepcopy(data)
    for package, info in packages_data.items():
        repo_name = info["repo_name"]
        authors = info.get("authors")
        if authors:
            for author in authors:
                author_login = author.get("login", "No_author_login")
                commit_sha = author.get("sha", "No_commit_sha")
                author_type = author.get("a_type", "No_author_type")

                commit_result = {
                    "api_url": None,
                    "earliest_commit_sha": None,
                    "author_login_in_1st_commit": None,
                    "author_id_in_1st_commit": None,
                    "is_first_commit": None,
                    "commit_notice": None,
                }

                if author_login is None:
                    commit_result["commit_notice"] = "Author login is None"

                if "[bot]" in author_login or author_type == "Bot":
                    commit_result["earliest_commit_sha"] = "It might be a bot"
                    commit_result["commit_notice"] = "Bot author detected"
                else:
                    api_url = f"https://api.github.com/repos/{repo_name}/commits?author={author_login}"
                    data = cache_manager.user_commit_cache.get_user_commit(api_url)
                    if data:
                        # Retrieved data from cache
                        (
                            earliest_commit_sha,
                            author_login_in_commit,
                            author_id_in_commit,
                        ) = data
                        first_time_commit = earliest_commit_sha == commit_sha

                        commit_result.update(
                            {
                                "api_url": api_url,
                                "earliest_commit_sha": earliest_commit_sha,
                                "author_login_in_1st_commit": author_login_in_commit,
                                "author_id_in_1st_commit": author_id_in_commit,
                                "is_first_commit": first_time_commit,
                                "commit_notice": "Data retrieved from cache",
                            }
                        )
                    else:
                        # Data not found in cache, need to make API request
                        result = get_repo_author_commits(api_url)
                        if result:
                            earliest_commit = result[0]
                            earliest_commit_sha = earliest_commit["sha"]
                            author_login_in_commit = earliest_commit["author"]["login"]
                            author_id_in_commit = earliest_commit["author"]["id"]
                            first_time_commit = earliest_commit_sha == commit_sha
                            cache_manager.user_commit_cache.cache_user_commit(
                                api_url,
                                earliest_commit_sha,
                                repo_name,
                                package,
                                author_login,
                                commit_sha,
                                author_login_in_commit,
                                author_id_in_commit,
                            )
                            commit_result.update(
                                {
                                    "api_url": api_url,
                                    "earliest_commit_sha": earliest_commit_sha,
                                    "author_login_in_1st_commit": author_login_in_commit,
                                    "author_id_in_1st_commit": author_id_in_commit,
                                    "is_first_commit": first_time_commit,
                                    "commit_notice": "Data retrieved from API",
                                }
                            )
                        else:
                            commit_result["commit_notice"] = f"Failed to retrieve data from API({api_url})"
                author["commit_result"] = commit_result
        else:
            info["commit_result"] = None

    return packages_data



"""
dirty-waters - A tool to detect software supply chain smells and issues
"""

from .main import main
from .report_static import write_summary
from .report_diff import generate_diff_report

__version__ = "0.1.1"
__author__ = "CHAINS research project at KTH Royal Institute of Technology"

__all__ = ["main", "write_summary", "generate_diff_report"]



import os
import logging
from typing import List, Dict, Tuple
from tool.tool_config import get_cache_manager, make_github_request

cache_manager = get_cache_manager()

GITHUB_TOKEN = os.getenv("GITHUB_API_TOKEN")
BATCH_SIZE = 100  # Configurable batch size

headers = {
    "Authorization": f"Bearer {GITHUB_TOKEN}",
    "Accept": "application/vnd.github.v4+json",
}

url = "https://api.github.com/graphql"


def fetch_and_cache_batch(commit_batch: List[Tuple[str, str, str, str]]) -> List[Dict]:
    """
    Fetch and cache PR information for a batch of commits.

    Args:
        commit_batch: List of tuples (node_id, commit_sha, package, repo_name)

    Returns:
        List of processed PR information dictionaries
    """
    if not commit_batch:
        return []

    # Build the GraphQL query for this batch
    query_parts = []
    variables = {"first": 5}  # Number of PRs to fetch per commit

    for i, (node_id, _, _, _) in enumerate(commit_batch):
        variables[f"nodeId{i}"] = node_id
        query_parts.append(
            f"""
        node{i}: node(id: $nodeId{i}) {{
            ... on Commit {{
                associatedPullRequests(first: $first) {{
                    edges {{
                        node {{
                            author {{
                                login
                                __typename
                            }}
                            authorAssociation
                            autoMergeRequest {{
                                mergeMethod
                                enabledBy {{
                                    login
                                }}
                                authorEmail
                            }}
                            checksUrl
                            createdAt
                            mergeCommit {{
                                author {{
                                    name
                                    email
                                }}
                            }}
                            id
                            merged
                            mergedAt
                            mergedBy {{
                                login
                                __typename
                            }}
                            number
                            state
                            url
                            reviews(first: $first, states: APPROVED) {{
                                edges {{
                                    node {{
                                        author {{
                                            login
                                            __typename
                                        }}
                                        id
                                        state
                                        createdAt
                                        publishedAt
                                        submittedAt
                                        updatedAt
                                    }}
                                }}
                            }}
                            repository {{
                                name
                                owner {{
                                    login
                                    id
                                }}
                            }}
                        }}
                    }}
                }}
            }}
        }}
        """
        )

    # Construct the full query
    query = (
        "query("
        + ", ".join([f"$nodeId{i}: ID!" for i in range(len(commit_batch))])
        + ", $first: Int) {"
        + "\n".join(query_parts)
        + "}"
    )

    # Execute the query
    body = {"query": query, "variables": variables}
    response = make_github_request(url, method="POST", json_data=body, headers=headers, max_retries=3)

    batch_results = []
    if response and "data" in response:
        # Process and cache each result immediately
        for i, (node_id, commit_sha, package, repo_name) in enumerate(commit_batch):
            node_key = f"node{i}"
            pr_info = {}
            if (
                node_key in response["data"]
                and response["data"][node_key] is not None
                and "associatedPullRequests" in response["data"][node_key]
            ):
                pr_info = {"data": {"node": response["data"][node_key]}}

            # Cache immediately after processing each item
            cache_manager.github_cache.cache_pr_info(
                {
                    "package": package,
                    "commit_sha": commit_sha,
                    "commit_node_id": node_id,
                    "pr_info": pr_info,
                }
            )

            # Add to batch results
            batch_results.append(
                {
                    "package": package,
                    "commit_sha": commit_sha,
                    "commit_node_id": node_id,
                    "pr_info": pr_info,
                    "repo_name": repo_name,
                }
            )

            logging.info(f"Processed and cached PR info for commit {commit_sha} in {package}")
    else:
        # Handle error case
        logging.error(f"Failed to fetch PR information for batch of size {len(commit_batch)}")
        logging.error(f"Response: {response}")
        # Cache empty results for failed requests to prevent repeated failures
        for node_id, commit_sha, package, repo_name in commit_batch:
            cache_manager.github_cache.cache_pr_info(
                {
                    "package": package,
                    "commit_sha": commit_sha,
                    "commit_node_id": node_id,
                    "pr_info": {},
                }
            )
            batch_results.append(
                {
                    "package": package,
                    "commit_sha": commit_sha,
                    "commit_node_id": node_id,
                    "pr_info": {},
                    "repo_name": repo_name,
                }
            )

    return batch_results


def get_pr_info(data: Dict) -> List[Dict]:
    """
    Get PR information for all commits, processing in batches and caching gradually.

    Args:
        data: Dictionary containing commit information by package

    Returns:
        List of PR information dictionaries
    """
    logging.info("Getting PR info for commits...")

    pr_infos = []
    commits_to_process = []

    # First pass: collect commits that need processing
    for package, info in data.items():
        repo_name = info.get("repo_name")
        authors = info.get("authors", [])

        for author in authors:
            commit_sha = author.get("sha")
            commit_node_id = author.get("node_id")

            if not commit_node_id:
                continue

            # Check cache first
            pr_data = cache_manager.github_cache.get_pr_info(commit_node_id)
            if pr_data:
                # Use cached data
                pr_infos.append(
                    {
                        "package": package,
                        "commit_sha": commit_sha,
                        "commit_node_id": commit_node_id,
                        "pr_info": pr_data["pr_info"],
                        "repo_name": repo_name,
                    }
                )
            else:
                # Add to list for batch processing
                commits_to_process.append((commit_node_id, commit_sha, package, repo_name))

    # Process commits in batches
    total_commits = len(commits_to_process)
    if total_commits > 0:
        logging.info(f"Processing {total_commits} commits in batches of {BATCH_SIZE}")

        for i in range(0, total_commits, BATCH_SIZE):
            batch = commits_to_process[i : i + BATCH_SIZE]
            logging.info(f"Processing batch {i//BATCH_SIZE + 1}/{(total_commits + BATCH_SIZE - 1)//BATCH_SIZE}")

            try:
                batch_results = fetch_and_cache_batch(batch)
                pr_infos.extend(batch_results)

                # Log progress
                processed = min(i + BATCH_SIZE, total_commits)
                logging.info(f"Processed {processed}/{total_commits} commits ({processed/total_commits*100:.1f}%)")

            except Exception as e:
                logging.error(f"Error processing batch: {e}")
                # Continue with next batch instead of failing completely
                continue

    return pr_infos


def get_useful_pr_info(commits_data):
    pr_infos = get_pr_info(commits_data)

    for pr_info in pr_infos:
        if pr_info:
            package = pr_info.get("package")
            commit_sha = pr_info.get("commit_sha")
            commit_node_id = pr_info.get("commit_node_id")
            repo_name = pr_info.get("repo_name")
            associated_prs = (
                pr_info.get("pr_info", {})
                .get("data", {})
                .get("node", {})
                .get("associatedPullRequests", {})
                .get("edges", [])
            )
            for author in commits_data[package].get("authors", []):
                if author.get("node_id") == commit_node_id:
                    author["commit_merged_info"] = []
                    if len(associated_prs) == 0:
                        author["commit_merged_info"].append({"merge_info": "no associated PRs"})
                    else:
                        for associated_pr in associated_prs:
                            merge_node_info = associated_pr.get("node", {})
                            author_association = merge_node_info.get("authorAssociation")
                            auto_merge_request = merge_node_info.get("autoMergeRequest", {})
                            created_at = merge_node_info.get("createdAt")
                            pr_id = merge_node_info.get("id")
                            state = merge_node_info.get("state")
                            merge_at = merge_node_info.get("mergedAt")
                            pull_url = merge_node_info.get("url")
                            reviews = merge_node_info.get("reviews", {}).get("edges", [])

                            merged_info = {
                                "repo": repo_name,
                                "commit_sha": commit_sha,
                                "commit_node_id": commit_node_id,
                                "author_association": author_association,
                                "auto_merge_request": auto_merge_request,
                                "created_at": created_at,
                                "pr_id": pr_id,
                                "state": state,
                                "merge_at": merge_at,
                                "pull_url": pull_url,
                                "reviews": [],
                            }

                            if state == "MERGED":
                                if merge_node_info.get("mergedBy"):
                                    merged_info["merge_by"] = merge_node_info.get("mergedBy", {}).get("login")
                                    merged_info["merge_by_type"] = merge_node_info.get("mergedBy", {}).get(
                                        "__typename"
                                    )
                                else:
                                    merged_info["merge_by"] = merge_node_info.get("mergedBy")

                            else:
                                merged_info["merge_by"] = None
                                merged_info["merge_by_type"] = None

                            for review in reviews:
                                review_node = review.get("node", {})
                                if review_node:
                                    if review_node.get("author", {}):
                                        review_author = review_node.get("author", {}).get("login", None)
                                        review_author_type = review_node.get("author", {}).get("__typename", None)

                                        review_info = {
                                            "review_author": review_author,
                                            "review_author_type": review_author_type,
                                            "review_state": review_node.get("state", None),
                                            "review_id": review_node.get("id", None),
                                        }

                                        merged_info["reviews"].append(review_info)

                            author["commit_merged_info"].append(merged_info)

    return commits_data



"""
This file contains the configuration for the tool.
"""

import pathlib
import logging
import os
import requests_cache
import requests
import sqlite3
import json
from datetime import datetime, timedelta
from functools import lru_cache
from pathlib import Path
from typing import Dict, Optional, Union, List
import time
from git import Repo
import hashlib
import re

PNPM_LIST_COMMAND = lambda scope: [
    "pnpm",
    "list",
    "--filter",
    scope,
    "--depth",
    "Infinity",
]

github_token = os.getenv("GITHUB_API_TOKEN")

headers = {
    "Authorization": f"Bearer {github_token}",
    "Accept": "application/vnd.github.v3+json",
}


class PathManager:
    """
    Manage the paths for the results.
    """

    def __init__(self, base_dir="results"):
        self.base_dir = pathlib.Path(base_dir)

    def create_folders(self, version_tag):
        """
        Create the folders for the results.
        """

        current_time = datetime.now().strftime("%Y-%m-%d-%H-%M-%S")
        folder_name = f"results_{current_time}"
        result_folder_path = self.base_dir / folder_name
        result_folder_path.mkdir(parents=True, exist_ok=True)

        json_directory = result_folder_path / "sscs" / version_tag
        json_directory.mkdir(parents=True, exist_ok=True)
        diff_directory = result_folder_path / "diff"
        diff_directory.mkdir(parents=True, exist_ok=True)

        return result_folder_path, json_directory, diff_directory


class CacheManager:
    def __init__(self, cache_dir="cache"):
        self.cache_dir = Path(cache_dir)
        self.cache_dir.mkdir(parents=True, exist_ok=True)

        # Initialize all cache instances
        self.github_cache = GitHubCache(cache_dir)
        self.package_cache = PackageAnalysisCache(cache_dir)
        self.commit_comparison_cache = CommitComparisonCache(cache_dir)
        self.user_commit_cache = UserCommitCache(cache_dir)
        self.extracted_deps_cache = DependencyExtractionCache(cache_dir)

    def _setup_requests_cache(self, cache_name="http_cache"):
        requests_cache.install_cache(
            cache_name=str(self.cache_dir / f"{cache_name}_cache"),
            backend="sqlite",
            expire_after=7776000,  # 90 days
            allowable_codes=(200, 301, 302, 404),
        )

    def clear_all_caches(self, older_than_days=None):
        """Clear all caches"""
        self.github_cache.clear_cache(older_than_days)
        self.package_cache.clear_cache(older_than_days)
        self.commit_comparison_cache.clear_cache(older_than_days)
        self.user_commit_cache.clear_cache(older_than_days)
        self.extracted_deps_cache.clear_cache(older_than_days)


class Cache:
    def __init__(self, cache_dir="cache", db_name="cache.db"):
        self.cache_dir = Path(cache_dir)
        self.cache_dir.mkdir(parents=True, exist_ok=True)
        self.db_path = self.cache_dir / db_name
        self._execute_query(
            """
            CREATE TABLE IF NOT EXISTS schema_signatures (
                table_name TEXT PRIMARY KEY,
                signature TEXT,
                last_updated TIMESTAMP
            )
        """
        )
        self.setup_db()

    def setup_db(self):
        """Initialize SQLite database - should be implemented by subclasses"""
        raise NotImplementedError

    def _execute_query(self, query, params=None):
        """Execute SQLite query with proper connection handling"""
        conn = sqlite3.connect(self.db_path)
        c = conn.cursor()
        try:
            if params:
                c.execute(query, params)
            else:
                c.execute(query)
            conn.commit()
            return c.fetchall()
        finally:
            conn.close()

    def _generate_schema_signature(self, schema):
        """Generate a unique signature for a schema definition"""
        # Removing whitespace and convert to lowercase to ignore formatting differences
        normalized_schema = " ".join(schema.lower().split())
        return hashlib.md5(normalized_schema.encode()).hexdigest()

    def _check_and_update_table(self, table_name, schema):
        """Check if table exists with correct schema, otherwise recreate it"""
        new_signature = self._generate_schema_signature(schema)
        current_signature = self._get_table_signature(table_name)

        if current_signature is None:
            # Table doesn't exist or isn't versioned yet
            print(f"Creating new table: {table_name}")
            self._create_new_table(table_name, schema, new_signature)
        elif current_signature[0] != new_signature:
            # Table exists but schema has changed
            print(f"Updating table: {table_name}")
            self._update_table(table_name, schema, new_signature)
        else:
            print(f"Table {table_name} is up to date")

    def _get_table_signature(self, table_name):
        """Get the current signature of a table from schema_signatures"""
        try:
            result = self._execute_query("SELECT signature FROM schema_signatures WHERE table_name = ?", (table_name,))
            return result[0] if result else None
        except:
            # Table might not exist yet
            return None

    def _create_new_table(self, table_name, schema, signature):
        """Create a new table and record its signature"""
        # Check if table exists already (but isn't tracked)
        table_exists = self._check_table_exists(table_name)

        if table_exists:
            self._execute_query(f"DROP TABLE {table_name}")
        self._execute_query(schema)
        self._execute_query(
            """
            INSERT INTO schema_signatures (table_name, signature, last_updated)
            VALUES (?, ?, CURRENT_TIMESTAMP)
            """,
            (table_name, signature),
        )

    def _update_table(self, table_name, schema, new_signature):
        """Update an existing table to a new schema"""
        self._execute_query(f"DROP TABLE {table_name}")
        self._execute_query(schema)
        self._execute_query(
            """
            UPDATE schema_signatures 
            SET signature = ?, last_updated = CURRENT_TIMESTAMP
            WHERE table_name = ?
            """,
            (new_signature, table_name),
        )

    def _check_table_exists(self, table_name):
        """Check if a table exists in the database"""
        try:
            result = self._execute_query(
                """
                SELECT name FROM sqlite_master 
                WHERE type='table' AND name=?
                """,
                (table_name,),
            )
            return result not in [None, []]
        except:
            return False

    def clear_cache(self, older_than_days=None):
        """Clear cached data older than specified days"""
        if older_than_days:
            cutoff = (datetime.now() - timedelta(days=older_than_days)).isoformat()
            self._execute_query("DELETE FROM cache_entries WHERE cached_at < ?", (cutoff,))
        else:
            self._execute_query("DELETE FROM cache_entries")


class GitHubCache(Cache):
    def __init__(self, cache_dir="cache/github"):
        super().__init__(cache_dir, "github_cache.db")
        self.repo_cache = {}  # In-memory LRU cache

    def setup_db(self):
        """Initialize GitHub-specific cache tables with automatic schema versioning"""
        table_schemas = {
            "github_urls": """
                CREATE TABLE github_urls (
                    package TEXT PRIMARY KEY,
                    repo_url TEXT,
                    cached_at TIMESTAMP
                )
            """,
            "pr_info": """
                CREATE TABLE pr_info (
                    package TEXT,
                    commit_sha TEXT,
                    commit_node_id TEXT PRIMARY KEY,
                    pr_info TEXT,
                    cached_at TIMESTAMP
                )
            """,
            "pr_reviews": """
                CREATE TABLE pr_reviews (
                    package TEXT,
                    repo_name TEXT,
                    author TEXT,
                    first_review_data TEXT,
                    cached_at TIMESTAMP,
                    PRIMARY KEY (repo_name, author)
                )
            """,
            "tag_to_sha": """
                CREATE TABLE tag_to_sha (
                    repo_name TEXT,
                    tag TEXT,
                    sha TEXT,
                    cached_at TIMESTAMP,
                    PRIMARY KEY (repo_name, tag)
                )
            """,
        }

        for table_name, schema in table_schemas.items():
            self._check_and_update_table(table_name, schema)

    def cache_pr_review(self, package, repo_name, author, first_review_data):
        """Cache PR review information"""
        conn = sqlite3.connect(self.db_path)
        c = conn.cursor()

        try:
            c.execute(
                """
                INSERT OR REPLACE INTO pr_reviews
                (package, repo_name, author, first_review_data, cached_at)
                VALUES (?, ?, ?, ?, ?)
            """,
                (package, repo_name, author, json.dumps(first_review_data), datetime.now().isoformat()),
            )
            conn.commit()
        finally:
            conn.close()

    def get_pr_review(self, repo_name=None, author=None):
        """Get PR review information from cache"""
        conn = sqlite3.connect(self.db_path)
        c = conn.cursor()

        try:
            c.execute(
                "SELECT first_review_data, cached_at FROM pr_reviews WHERE repo_name = ? AND author = ?",
                (repo_name, author),
            )
            result = c.fetchone()
            if result:
                review_data, cached_at = result
                cached_at = datetime.fromisoformat(cached_at)

                # Return cached data if it's less than 30 days old
                if datetime.now() - cached_at < timedelta(days=30):
                    return json.loads(review_data)
            return None
        finally:
            conn.close()

    def cache_github_url(self, package, repo_info):
        """Cache GitHub URL for a package"""
        conn = sqlite3.connect(self.db_path)
        c = conn.cursor()

        try:
            c.execute(
                """
                INSERT OR REPLACE INTO github_urls 
                (package, repo_url, cached_at)
                VALUES (?, ?, ?)
            """,
                (package, json.dumps(repo_info), datetime.now().isoformat()),
            )
            conn.commit()
        finally:
            conn.close()

    def get_github_url(self, package):
        """Get cached GitHub URL for a package"""
        conn = sqlite3.connect(self.db_path)
        c = conn.cursor()

        try:
            c.execute("SELECT repo_url, cached_at FROM github_urls WHERE package = ?", (package,))
            result = c.fetchone()

            if result:
                repo_info, cached_at = result
                cached_at = datetime.fromisoformat(cached_at)

                # URLs don't change often, so we can cache them for longer (180 days)
                if datetime.now() - cached_at < timedelta(days=180):
                    return json.loads(repo_info)

            return None
        finally:
            conn.close()

    def cache_pr_info(self, pr_data: Dict):
        """Cache PR info with current timestamp"""
        with sqlite3.connect(self.db_path) as conn:
            conn.execute(
                """
                INSERT OR REPLACE INTO pr_info
                (package, commit_sha, commit_node_id, pr_info, cached_at)
                VALUES (?, ?, ?, ?, ?)
            """,
                (
                    pr_data["package"],
                    pr_data["commit_sha"],
                    pr_data["commit_node_id"],
                    json.dumps(pr_data["pr_info"]),
                    datetime.now().isoformat(),
                ),
            )
            conn.commit()

    def get_pr_info(self, commit_node_id: str) -> Optional[Dict]:
        """Get PR info from cache if available and not expired"""
        conn = sqlite3.connect(self.db_path)
        c = conn.cursor()
        with sqlite3.connect(self.db_path) as conn:
            c.execute(
                "SELECT package, commit_sha, commit_node_id, pr_info, cached_at FROM pr_info WHERE commit_node_id = ?",
                (commit_node_id,),
            )
            result = c.fetchone()

            if result:
                package, commit_sha, commit_node_id, pr_info, cached_at = result
                cached_at = datetime.fromisoformat(cached_at)
                if datetime.now() - cached_at < timedelta(hours=24):
                    return {
                        "package": package,
                        "commit_sha": commit_sha,
                        "commit_node_id": commit_node_id,
                        "pr_info": json.loads(pr_info),
                    }
        return None

    def cache_tag_to_sha(self, repo_name, tag, sha):
        """Cache tag to SHA mapping"""
        with sqlite3.connect(self.db_path) as conn:
            conn.execute(
                """
                INSERT OR REPLACE INTO tag_to_sha
                (repo_name, tag, sha, cached_at)
                VALUES (?, ?, ?, ?)
            """,
                (repo_name, tag, sha, datetime.now().isoformat()),
            )
            conn.commit()

    def get_tag_to_sha(self, repo_name, tag):
        """Get SHA for a tag from cache"""
        with sqlite3.connect(self.db_path) as conn:
            c = conn.cursor()
            c.execute("SELECT sha, cached_at FROM tag_to_sha WHERE repo_name = ? AND tag = ?", (repo_name, tag))
            result = c.fetchone()

            if result:
                sha, cached_at = result
                cached_at = datetime.fromisoformat(cached_at)
                if datetime.now() - cached_at < timedelta(days=180):
                    return sha
        return None

    def clear_cache(self, older_than_days=None):
        """Clear cached data"""
        conn = sqlite3.connect(self.db_path)
        c = conn.cursor()

        try:
            if older_than_days:
                cutoff = (datetime.now() - timedelta(days=older_than_days)).isoformat()
                c.execute("DELETE FROM github_urls WHERE cached_at < ?", (cutoff,))
                c.execute("DELETE FROM pr_info WHERE cached_at < ?", (cutoff,))
                c.execute("DELETE FROM pr_reviews WHERE cached_at < ?", (cutoff,))
                c.execute("DELETE FROM tag_to_sha WHERE cached_at < ?", (cutoff,))
            else:
                c.execute("DELETE FROM github_urls")
                c.execute("DELETE FROM pr_info")
                c.execute("DELETE FROM pr_reviews")
                c.execute("DELETE FROM tag_to_sha")
            conn.commit()

        finally:
            conn.close()

    def clear_github_urls_from_package(self, package):
        """Clear cached GitHub URLs for a package"""
        conn = sqlite3.connect(self.db_path)
        c = conn.cursor()

        try:
            # first get count of rows with the package name
            c.execute("SELECT COUNT(*) FROM github_urls WHERE package = ?", (package,))
            count = c.fetchone()[0]
            if count == 0:
                print(f"No cached data found for {package}")
                logging.info(f"No cached data found for {package}")
                return

            # delete rows with the package name
            print(f"Deleting cached data for {package}")
            logging.info(f"Deleting cached data for {package}")
            c.execute("DELETE FROM github_urls WHERE package = ?", (package,))
            conn.commit()
        finally:
            conn.close()


class PackageAnalysisCache(Cache):
    def __init__(self, cache_dir="cache/packages"):
        super().__init__(cache_dir, "package_analysis.db")

    def setup_db(self):
        """Initialize package analysis cache tables with automatic schema versioning"""
        table_schemas = {
            "package_analysis": """
                CREATE TABLE package_analysis (
                    package_name TEXT,
                    version TEXT,
                    package_manager TEXT,
                    analysis_data TEXT,
                    cached_at TIMESTAMP,
                    PRIMARY KEY (package_name, version, package_manager)
                )
            """
        }

        for table_name, schema in table_schemas.items():
            self._check_and_update_table(table_name, schema)

    def cache_package_analysis(self, package_name, version, package_manager, analysis_data):
        """Cache package analysis results"""
        self._execute_query(
            """
            INSERT OR REPLACE INTO package_analysis 
            (package_name, version, package_manager, analysis_data, cached_at)
            VALUES (?, ?, ?, ?, ?)
        """,
            (package_name, version, package_manager, json.dumps(analysis_data), datetime.now().isoformat()),
        )

    def get_package_analysis(self, package_name, version, package_manager, max_age_days=180):
        """Get cached package analysis results"""
        results = self._execute_query(
            """SELECT analysis_data, cached_at 
               FROM package_analysis 
               WHERE package_name = ? AND version = ? AND package_manager = ?""",
            (package_name, version, package_manager),
        )

        if results:
            analysis_data, cached_at = results[0]
            cached_at = datetime.fromisoformat(cached_at)

            if datetime.now() - cached_at < timedelta(days=max_age_days):
                return json.loads(analysis_data)

        return None

    def clear_cache(self, older_than_days=None):
        """Clear cached data"""
        conn = sqlite3.connect(self.db_path)
        c = conn.cursor()

        try:
            if older_than_days:
                cutoff = (datetime.now() - timedelta(days=older_than_days)).isoformat()
                c.execute("DELETE FROM package_analysis WHERE cached_at < ?", (cutoff,))
            else:
                c.execute("DELETE FROM package_analysis")

            conn.commit()

        finally:
            conn.close()

    def clear_package_by_version(self, package_name, version):
        """Clear cached data for a specific package version"""
        conn = sqlite3.connect(self.db_path)
        c = conn.cursor()

        try:
            c.execute(
                "SELECT COUNT(*) FROM package_analysis WHERE package_name = ? AND version = ?",
                (package_name, version),
            )
            count = c.fetchone()[0]
            if count == 0:
                print(f"No cached data found for {package_name} {version}")
                logging.info(f"No cached data found for {package_name} {version}")
                return

            c.execute("DELETE FROM package_analysis WHERE package_name = ? AND version = ?", (package_name, version))
            conn.commit()
            print(f"Cleared cached data for {package_name} {version}")
            logging.info(f"Cleared cached data for {package_name} {version}")

        finally:
            conn.close()


class CommitComparisonCache(Cache):
    def __init__(self, cache_dir="cache/commits"):
        super().__init__(cache_dir, "commit_comparison_cache.db")

    def setup_db(self):
        """Initialize commit comparison cache tables with automatic schema versioning"""
        table_schemas = {
            "commit_authors_from_tags": """
                CREATE TABLE commit_authors_from_tags (
                    package TEXT,
                    tag1 TEXT,
                    tag2 TEXT,
                    data TEXT,
                    cached_at TIMESTAMP,
                    PRIMARY KEY (package, tag1, tag2)
                )
            """,
            "commit_authors_from_url": """
                CREATE TABLE commit_authors_from_url (
                    commit_url TEXT PRIMARY KEY,
                    data TEXT,
                    cached_at TIMESTAMP
                )
            """,
            "patch_authors_from_sha": """
                CREATE TABLE patch_authors_from_sha (
                    repo_name TEXT,
                    patch_path TEXT,
                    sha TEXT,
                    data TEXT,
                    cached_at TIMESTAMP,
                    PRIMARY KEY (repo_name, patch_path, sha)
                )
            """,
        }

        for table_name, schema in table_schemas.items():
            self._check_and_update_table(table_name, schema)

    def cache_authors_from_tags(self, package, tag1, tag2, data):
        self._execute_query(
            """
            INSERT OR REPLACE INTO commit_authors_from_tags 
            (package, tag1, tag2, data, cached_at)
            VALUES (?, ?, ?, ?, ?)
        """,
            (package, tag1, tag2, json.dumps(data), datetime.now().isoformat()),
        )

    def get_authors_from_tags(self, package, tag1, tag2, max_age_days=180):
        results = self._execute_query(
            "SELECT data, cached_at FROM commit_authors_from_tags WHERE package = ? AND tag1 = ? AND tag2 = ?",
            (package, tag1, tag2),
        )
        if results:
            data, cached_at = results[0]
            cached_at = datetime.fromisoformat(cached_at)
            if datetime.now() - cached_at < timedelta(days=max_age_days):
                return json.loads(data)
        return None

    def cache_authors_from_url(self, commit_url, data):
        self._execute_query(
            """
            INSERT OR REPLACE INTO commit_authors_from_url 
            (commit_url, data, cached_at)
            VALUES (?, ?, ?)
        """,
            (commit_url, json.dumps(data), datetime.now().isoformat()),
        )

    def get_authors_from_url(self, commit_url, max_age_days=180):
        results = self._execute_query(
            "SELECT data, cached_at FROM commit_authors_from_url WHERE commit_url = ?", (commit_url,)
        )
        if results:
            data, cached_at = results[0]
            cached_at = datetime.fromisoformat(cached_at)
            if datetime.now() - cached_at < timedelta(days=max_age_days):
                return json.loads(data)
        return None

    def cache_patch_authors(self, repo_name, patch_path, sha, data):
        self._execute_query(
            """
            INSERT OR REPLACE INTO patch_authors_from_sha 
            (repo_name, patch_path, sha, data, cached_at)
            VALUES (?, ?, ?, ?, ?)
        """,
            (repo_name, patch_path, sha, json.dumps(data), datetime.now().isoformat()),
        )

    def get_patch_authors(self, repo_name, patch_path, sha, max_age_days=180):
        results = self._execute_query(
            "SELECT data, cached_at FROM patch_authors_from_sha WHERE repo_name = ? AND patch_path = ? AND sha = ?",
            (repo_name, patch_path, sha),
        )
        if results:
            data, cached_at = results[0]
            cached_at = datetime.fromisoformat(cached_at)
            if datetime.now() - cached_at < timedelta(days=max_age_days):
                return json.loads(data)
        return None

    def clear_cache(self, older_than_days=None):
        """Clear cached data"""
        conn = sqlite3.connect(self.db_path)
        c = conn.cursor()

        try:
            if older_than_days:
                cutoff = (datetime.now() - timedelta(days=older_than_days)).isoformat()
                c.execute("DELETE FROM commit_authors_from_tags WHERE cached_at < ?", (cutoff,))
                c.execute("DELETE FROM commit_authors_from_url WHERE cached_at < ?", (cutoff,))
                c.execute("DELETE FROM patch_authors_from_sha WHERE cached_at < ?", (cutoff,))
            else:
                c.execute("DELETE FROM commit_authors_from_tags")
                c.execute("DELETE FROM commit_authors_from_url")
                c.execute("DELETE FROM patch_authors_from_sha")

            conn.commit()

        finally:
            conn.close()


class UserCommitCache(Cache):
    def __init__(self, cache_dir="cache/user_commits"):
        super().__init__(cache_dir, "user_commits.db")

    def setup_db(self):
        """Initialize user commit cache tables with automatic schema versioning"""
        table_schemas = {
            "user_commit": """
                CREATE TABLE user_commit (
                    api_url TEXT PRIMARY KEY,
                    earliest_commit_sha TEXT,
                    repo_name TEXT,
                    package TEXT,
                    author_login TEXT,
                    author_commit_sha TEXT,
                    author_login_in_1st_commit TEXT,
                    author_id_in_1st_commit TEXT,
                    cached_at TIMESTAMP
                )
            """
        }

        for table_name, schema in table_schemas.items():
            self._check_and_update_table(table_name, schema)

    def cache_user_commit(
        self,
        api_url,
        earliest_commit_sha,
        repo_name,
        package,
        author_login,
        author_commit_sha,
        author_login_in_1st_commit,
        author_id_in_1st_commit,
    ):
        self._execute_query(
            """
            INSERT OR REPLACE INTO user_commit 
            (api_url, earliest_commit_sha, repo_name, package, author_login, author_commit_sha, author_login_in_1st_commit, author_id_in_1st_commit, cached_at)
            VALUES (?, ?, ?, ?, ?, ?, ?, ?, ?)
        """,
            (
                api_url,
                earliest_commit_sha,
                repo_name,
                package,
                author_login,
                author_commit_sha,
                author_login_in_1st_commit,
                author_id_in_1st_commit,
                datetime.now().isoformat(),
            ),
        )

    def get_user_commit(self, api_url, max_age_days=180):
        results = self._execute_query(
            "SELECT earliest_commit_sha, author_login_in_1st_commit, author_id_in_1st_commit, cached_at FROM user_commit WHERE api_url = ?",
            (api_url,),
        )
        if results:
            earliest_commit_sha, author_login_in_1st_commit, author_id_in_1st_commit, cached_at = results[0]
            cached_at = datetime.fromisoformat(cached_at)
            if datetime.now() - cached_at < timedelta(days=max_age_days):
                return earliest_commit_sha, author_login_in_1st_commit, author_id_in_1st_commit
        return None

    def clear_cache(self, older_than_days=None):
        """Clear cached data"""
        conn = sqlite3.connect(self.db_path)
        c = conn.cursor()

        try:
            if older_than_days:
                cutoff = (datetime.now() - timedelta(days=older_than_days)).isoformat()
                c.execute("DELETE FROM user_commit WHERE cached_at < ?", (cutoff,))
            else:
                c.execute("DELETE FROM user_commit")

            conn.commit()

        finally:
            conn.close()


class DependencyExtractionCache(Cache):
    def __init__(self, cache_dir="cache/extracted_deps"):
        super().__init__(cache_dir, "maven_deps.db")

    def setup_db(self):
        """Initialize dependency extraction cache tables with automatic schema versioning"""
        table_schemas = {
            "extracted_dependencies": """
                CREATE TABLE extracted_dependencies (
                    repo_path TEXT,
                    file_hash TEXT,
                    dependencies TEXT,
                    cached_at TIMESTAMP,
                    PRIMARY KEY (repo_path, file_hash)
                )
            """
        }

        for table_name, schema in table_schemas.items():
            self._check_and_update_table(table_name, schema)

    def cache_dependencies(self, repo_path, file_hash, dependencies):
        self._execute_query(
            """
            INSERT OR REPLACE INTO extracted_dependencies 
            (repo_path, file_hash, dependencies, cached_at)
            VALUES (?, ?, ?, ?)
        """,
            (repo_path, file_hash, json.dumps(dependencies), datetime.now().isoformat()),
        )

    def get_dependencies(self, repo_path, file_hash, max_age_days=180):
        results = self._execute_query(
            "SELECT dependencies, cached_at FROM extracted_dependencies WHERE repo_path = ? AND file_hash = ?",
            (repo_path, file_hash),
        )
        if results:
            deps_json, cached_at = results[0]
            cached_at = datetime.fromisoformat(cached_at)
            if datetime.now() - cached_at < timedelta(days=max_age_days):
                return json.loads(deps_json)
        return None

    def clear_cache(self, older_than_days=None):
        """Clear cached data"""
        conn = sqlite3.connect(self.db_path)
        c = conn.cursor()

        try:
            if older_than_days:
                cutoff = (datetime.now() - timedelta(days=older_than_days)).isoformat()
                c.execute("DELETE FROM extracted_dependencies WHERE cached_at < ?", (cutoff,))
            else:
                c.execute("DELETE FROM extracted_dependencies")

            conn.commit()

        finally:
            conn.close()


cache_manager = CacheManager()


class YarnLockParser:
    def __init__(self, content: str):
        """
        Initialize the Yarn.lock v1 parser with file content.

        :param content: Full content of the yarn.lock file
        """
        self.raw_content = content
        self.dependencies: Dict[str, Dict[str, Union[str, Dict[str, str]]]] = {}

    def parse(self) -> Dict[str, Dict[str, Union[str, Dict[str, str]]]]:
        """
        Parse the Yarn.lock v1 file content and extract dependency information.

        :return: Dictionary of parsed dependencies
        """
        # Reset dependencies for each parse
        self.dependencies = {}

        # Split the file into individual dependency blocks
        dependency_blocks = self._split_dependency_blocks(self.raw_content)

        # Parse each dependency block
        for block in dependency_blocks:
            parsed_block = self._parse_dependency_block(block)
            if parsed_block:
                name, details = parsed_block
                self.dependencies[name] = details

        return self.dependencies

    def _split_dependency_blocks(self, content: str) -> List[str]:
        """
        Split the file content into individual dependency blocks.

        :param content: Full content of the yarn.lock file
        :return: List of dependency block strings
        """
        # Yarn.lock v1 uses double newline as block separator
        blocks = re.split(r"\n\n", content.strip())
        return [block for block in blocks if block.strip()]

    def _parse_dependency_block(self, block: str) -> Optional[tuple]:
        """
        Parse an individual dependency block.

        :param block: A single dependency block
        :return: Tuple of (dependency name, dependency details) or None
        """
        lines = block.split("\n")

        # First line typically contains the name and version
        first_line = lines[0].strip()

        # Skip comments or empty lines
        if first_line.startswith("#") or not first_line:
            return None

        # Extract name and version
        name_version_match = re.match(r'^"?([^@"]+)@(?:npm:([^@]+)@)?(.+)"?:', first_line)
        if not name_version_match:
            return None

        alias = name_version_match.group(1)
        original_name = name_version_match.group(2) if name_version_match.group(2) else None
        version_constraint = name_version_match.group(3)

        # name, version_constraint = name_version_match.groups()

        # Initialize details dictionary
        details: Dict[str, Union[str, Dict[str, str]]] = {
            "original_name": original_name,
            "version_constraint": version_constraint,
            "resolved": None,
            "integrity": None,
            "dependencies": {},
        }

        # Track parsing state
        current_section = "metadata"
        current_dependency = None

        # Parse subsequent lines for additional metadata and nested dependencies
        for line in lines[1:]:
            line = line.strip()

            # Check for version constraint
            version_match = re.match(r'version "(.*)"', line)
            if version_match and current_section == "metadata":
                details["version_constraint"] = version_match.group(1)
                continue

            # Check for resolved URL
            resolved_match = re.match(r'resolved "(.*)"', line)
            if resolved_match and current_section == "metadata":
                details["resolved"] = resolved_match.group(1)
                continue

            # Check for integrity hash
            integrity_match = re.match(r'integrity "(.*)"', line)
            if integrity_match and current_section == "metadata":
                details["integrity"] = integrity_match.group(1)
                continue

            # Handle dependencies section
            if line.startswith("dependencies:"):
                current_section = "dependencies"
                continue

            # Parse nested dependencies
            if current_section == "dependencies":
                # Check if this is a new nested dependency
                nested_dep_match = re.match(r'^([^\s]+) "(.*)"', line)
                if nested_dep_match:
                    current_dependency = nested_dep_match.group(1)
                    dep_version = nested_dep_match.group(2)
                    details["dependencies"][current_dependency] = dep_version

        return (alias, details)

    def get_dependency(self, name: str) -> Optional[Dict[str, Union[str, Dict[str, str]]]]:
        """
        Retrieve details for a specific dependency.

        :param name: Name of the dependency
        :return: Dependency details or None
        """
        return self.dependencies.get(name)

    def list_dependencies(self) -> List[str]:
        """
        List all parsed dependencies.

        :return: List of dependency names
        """
        return list(self.dependencies.keys())


def get_cache_manager():
    return cache_manager


CLONE_OPTIONS = {
    "blobless": "--filter=blob:none",
}


def clone_repo(project_repo_name, release_version=None, blobless=False):
    """
    Clone the repository for the given project and release version.

    Args:
        project_repo_name (str): The name of the project repository.
        release_version (str): The release version of the project.
        blobless (bool): Whether to clone the repository without blobs.

    Returns:
        str: The path to the cloned repository.
    """

    repo_url = f"https://github.com/{project_repo_name}.git"

    # Clone to /tmp folder; if it is already cloned, an error will be raised
    try:
        options = [CLONE_OPTIONS["blobless"]] if blobless else []
        Repo.clone_from(repo_url, f"/tmp/{project_repo_name}", multi_options=options)
    except Exception as e:
        # If the repo is already cloned, just fetch the latest changes
        logging.info(f"Repo already cloned. Fetching the latest changes...")
        repo = Repo(f"/tmp/{project_repo_name}")

        # Fetch the latest changes
        repo.remotes.origin.fetch()
    # Checkout to the release version if provided
    if release_version:
        repo = Repo(f"/tmp/{project_repo_name}")
        repo.git.checkout(release_version)

    return f"/tmp/{project_repo_name}"


DEFAULT_CONFIG_PATH = ".dirty-waters.json"
DEFAULT_CONFIG = {"ignore": []}


def load_config(config_path=None):
    """
    Load configuration from a JSON file.

    Args:
        config_path (str): Path to config file. If None, looks for .dirty-waters.json in current directory

    Returns:
        dict: Configuration dictionary
    """
    if not config_path:
        logging.info(f"No config file provided, using default config path: {DEFAULT_CONFIG_PATH}")
        config_path = DEFAULT_CONFIG_PATH

    if not os.path.exists(config_path):
        logging.warning(f"Config file not found at {config_path}, using default config")
        return DEFAULT_CONFIG

    try:
        with open(config_path, "r") as f:
            config = json.load(f)
            logging.info(f"Found config file at {config_path}")
            return {**DEFAULT_CONFIG, **config}
    except Exception as e:
        logging.warning(f"Error loading config file: {str(e)}")
        return DEFAULT_CONFIG


def setup_logger(log_file_path, debug=False):
    """
    Setup the logger for the analysis.
    """

    class CustomFormatter(logging.Formatter):
        """Custom formatter, includes color coding for log levels."""

        grey = "\x1b[38;20m"
        green = "\x1b[38;2;0;200;0m"
        yellow = "\x1b[38;2;255;255;0m"
        red = "\x1b[38;2;255;0;0m"
        bold_red = "\x1b[1;31m"
        reset = "\x1b[0m"
        fmt = "%(asctime)s:%(name)s:%(levelname)s:%(message)s"

        FORMATS = {
            logging.DEBUG: grey + fmt + reset,
            logging.INFO: green + fmt + reset,
            logging.WARNING: yellow + fmt + reset,
            logging.ERROR: red + fmt + reset,
            logging.CRITICAL: bold_red + fmt + reset,
        }

        def format(self, record):
            log_fmt = self.FORMATS.get(record.levelno)
            formatter = logging.Formatter(log_fmt)
            return formatter.format(record)

    # Set up the logger
    logger = logging.getLogger()
    logger.setLevel(logging.INFO)

    # Create a console handler
    console_handler = logging.StreamHandler()
    console_handler.setLevel(logging.WARNING if not debug else logging.INFO)

    # Create a file handler
    file_handler = logging.FileHandler(log_file_path)
    file_handler.setLevel(logging.INFO)

    # Create a formatter and set it for both handlers
    formatter = CustomFormatter()
    console_handler.setFormatter(formatter)
    file_handler.setFormatter(formatter)

    # Add the handlers to the logger
    logger.addHandler(console_handler)
    logger.addHandler(file_handler)

    return logger


def make_github_request(
    url: str,
    method: str = "GET",
    headers: Dict = headers,
    json_data: Optional[Dict] = None,
    max_retries: int = 1,
    retry_delay: int = 2,
    timeout: int = 20,
    sleep_between_requests: int = 0,
    silent: bool = False,
) -> Optional[Dict]:
    """
    Make a HTTP request with retry logic and rate limiting handling.

    Args:
        url (str): HTTP URL
        method (str): HTTP method ("GET" or "POST")
        headers (Dict): Request headers
        json_data (Optional[Dict]): JSON payload for POST requests
        max_retries (int): Maximum number of retry attempts
        retry_delay (int): Base time to wait between retries in seconds
        timeout (int): Request timeout in seconds
        silent (bool): Whether to suppress error logging

    Returns:
        Optional[Dict]: JSON response or None if request failed
    """
    for attempt in range(max_retries):
        try:
            response = requests.request(method=method, url=url, headers=headers, json=json_data, timeout=timeout)
            response.raise_for_status()
            return response.json()

        except requests.exceptions.RequestException as e:
            time.sleep(sleep_between_requests)
            if isinstance(e, requests.exceptions.HTTPError) and (
                e.response.status_code in [429, 403] or "rate limit" in e.response.text.lower()
            ):
                if attempt == max_retries - 1:
                    if not silent:
                        logging.error(f"Failed after {max_retries} attempts due to rate limiting: {e}")
                    return None

                # Get rate limit reset time and wait
                reset_time = int(e.response.headers.get("X-RateLimit-Reset", 0))
                wait_time = max(reset_time - int(time.time()), 0) + 1
                if not silent:
                    logging.warning(f"Rate limit exceeded. Waiting {wait_time} seconds...")
                time.sleep(wait_time)
            else:
                # Handle other errors
                if not silent:
                    logging.warning(f"Request failed: {e}")
                if attempt == max_retries - 1:
                    if e.response.status_code in [
                        502,
                        504,
                    ]:  # timeout, sometimes happens when the request is too large (e.g., too many tags)
                        return 504
                    return None
                time.sleep(retry_delay * (attempt + 1))

    return None


def get_last_page_info(
    url: str, max_retries: int = 1, retry_delay: int = 2, sleep_between_requests: int = 0
) -> Optional[int]:
    """
    Get the last page number from the response headers.

    Args:
        url (str): URL to get the last page number
        max_retries (int): Maximum number
        retry_delay (int): Base time to wait between retries in seconds
        sleep_between_requests (int): Time to sleep between requests in seconds

    Returns:
        Optional[int]: Last page number or None if request failed
    """

    # We can't just use make_github_request here because we need to access the response headers
    for attempt in range(max_retries):
        try:
            response = requests.get(url, headers=headers)
            response.raise_for_status()
            if "last" in response.links:
                last_page = int(response.links["last"]["url"].split("=")[-1])
            else:
                # Otherwise, the last page is the first page too
                last_page = 1
            return last_page

        except requests.exceptions.RequestException as e:
            time.sleep(sleep_between_requests)
            if attempt == max_retries - 1:
                logging.error(f"Failed after {max_retries} attempts: {e}")
                return None
            time.sleep(retry_delay * (attempt + 1))



"""
Main file to run the software supply chain smell analysis.
"""

import json
import argparse
import logging
import os
import sys
import requests
from git import Repo

# Allows for tool to be recognized as a package to import from
sys.path.append(os.path.dirname(os.path.dirname(os.path.abspath(__file__))))

from tool import extract_deps
from tool import github_repo
from tool import static_analysis
from tool import compare_packages
from tool import compare_commits
from tool import get_user_commit_info
from tool import get_pr_info
from tool import get_pr_review
from tool import tool_config
from tool import report_static
from tool import report_diff

github_token = os.getenv("GITHUB_API_TOKEN")
if not github_token:
    raise ValueError("GitHub API token(GITHUB_API_TOKEN) is not set in the environment variables.")

headers = {
    "Authorization": f"Bearer {github_token}",
    "Accept": "application/vnd.github.v3+json",
}

cache_manager = tool_config.get_cache_manager()


def get_args():
    """
    Get command line arguments.

    Returns:
        argparse.Namespace: The parsed command line arguments.
    """
    parser = argparse.ArgumentParser()
    parser.add_argument(
        "-p",
        "--project-repo-name",
        required=True,
        help="Specify the project repository name. Example: MetaMask/metamask-extension",
    )
    parser.add_argument(
        "-v",
        "--release-version-old",
        required=False,
        default="HEAD",
        help="The old release tag of the project repository. Defaults to HEAD. Example: v10.0.0",
    )
    parser.add_argument(
        "-vn",
        "--release-version-new",
        help="The new release version of the project repository.",
    )
    parser.add_argument(
        "-d",
        "--differential-analysis",
        action="store_true",
        help="Run differential analysis and generate a markdown report of the project",
    )
    parser.add_argument(
        "-n",
        "--name-match",
        action="store_true",
        help="Compare the package names with the name in the in the package.json file. This option will slow down the execution time due to the API rate limit of code search.",
    )
    parser.add_argument(
        "-pm",
        "--package-manager",
        required=True,
        help="The package manager used in the project.",
        choices=["yarn-classic", "yarn-berry", "pnpm", "npm", "maven"],
    )
    # parser.add_argument(
    #     "--pnpm-scope",
    #     required=False,
    #     help="Extract dependencies from pnpm with a specific scope using 'pnpm list --filter <scope> --depth Infinity' command.",
    # )
    parser.add_argument(
        "--debug",
        action="store_true",
        help="Enable debug mode.",
    )
    parser.add_argument(
        "--config",
        help="Path to configuration file (JSON)",
        type=str,
        required=False,
    )

    # Gradual report group -- mutually exclusive, preserves backward compatibility for --no-gradual-report
    gradual_report_group = parser.add_mutually_exclusive_group()
    gradual_report_group.add_argument(
        "--gradual-report",
        dest="gradual_report",
        type=lambda x: str(x).lower() in ["true", "1", "yes", "y"],
        default=True,
        help="Enable/disable gradual reporting (default: true)",
    )
    gradual_report_group.add_argument(
        "--no-gradual-report",
        dest="gradual_report",
        action="store_false",
        help="Disable gradual reporting (deprecated, use --gradual-report=false instead)",
    )

    # Add new smell check arguments
    smell_group = parser.add_argument_group("smell checks")
    smell_group.add_argument(
        "--check-source-code",
        action="store_true",
        help="Check for dependencies with no link to source code repositories",
    )
    smell_group.add_argument(
        "--check-source-code-sha",
        action="store_true",
        help="Check for dependencies with no commit sha/tag for release",
    )
    smell_group.add_argument(
        "--check-deprecated",
        action="store_true",
        help="Check for deprecated dependencies",
    )
    smell_group.add_argument(
        "--check-forks",
        action="store_true",
        help="Check for dependencies that are forks",
    )
    smell_group.add_argument(
        "--check-provenance",
        action="store_true",
        help="Check for dependencies with no build attestation",
    )
    smell_group.add_argument(
        "--check-code-signature",
        action="store_true",
        help="Check for dependencies with missing/invalid code signature",
    )
    smell_group.add_argument(
        "--check-aliased-packages",
        action="store_true",
        help="Check for aliased packages",
    )

    arguments = parser.parse_args()

    return arguments


def get_lockfile(project_repo_name, release_version, package_manager):
    """
    Get the lockfile for the given project and release version.

    Args:
        project_repo_name (str): The name of the project repository.
        release_version (str): The release version of the project.
        package_manager (str): The package manager used in the project.

    Returns:
        str: The content of the lockfile or pom.xml.
        str: The default branch of the project.
        str: The name of the project repository.
    """

    LOOKING_FOR = {
        "yarn-classic": "yarn.lock",
        "yarn-berry": "yarn.lock",
        "pnpm": "pnpm-lock.yaml",
        "npm": "package-lock.json",
        "maven": "pom.xml",
    }

    cache_manager._setup_requests_cache(cache_name="get_lockfile")
    try:
        lockfile_name = LOOKING_FOR[package_manager]
        logging.info(f"Getting {lockfile_name} for {project_repo_name}@{release_version}")
        logging.info(f"Package manager: {package_manager}")
    except KeyError:
        logging.error("Invalid package manager or lack of lockfile: %s", package_manager)
        raise ValueError("Invalid package manager or lack of lockfile.")

    file_url = f"https://api.github.com/repos/{project_repo_name}/contents/{lockfile_name}?ref={release_version}"
    response = requests.get(file_url, headers=headers, timeout=20)

    if response.status_code == 200:
        data = response.json()
        download_url = data.get("download_url")
        lock_content = requests.get(download_url, timeout=60).text
        logging.info(f"Got the {lockfile_name} file from {download_url}.")
    else:
        logging.error(f"Failed to get {lockfile_name}.")
        raise ValueError(f"Failed to get {lockfile_name}.")

    repo_branch_api = f"https://api.github.com/repos/{project_repo_name}"
    repo_branch_response = requests.get(repo_branch_api, headers=headers, timeout=20)

    if repo_branch_response.status_code == 200:
        data = repo_branch_response.json()
        default_branch = data["default_branch"]
        logging.info("%s is the default branch of %s.", default_branch, project_repo_name)
    else:
        raise ValueError("Failed to get default branch")

    return lock_content, default_branch, project_repo_name


def get_deps(folder_path, project_repo_name, release_version, package_manager):
    """
    Get the dependencies for the given project and release version.

    Args:
        folder_path (str): The path to the project folder.
        project_repo_name (str): The name of the project repository.
        release_version (str): The release version of the project.
        package_manager (str): The package manager used in the project.
    """
    patches_info = None
    deps_list_all = None

    logging.info("Getting dependencies for %s@%s...", project_repo_name, release_version)

    # if it is a pnpm monorepo
    if package_manager == "pnpm":
        # if get_args().pnpm_scope:
        #     pnpm_scope = get_args().pnpm_scope
        #     deps_list_all = extract_deps.extract_deps_from_pnpm_mono(
        #         folder_path, release_version, project_repo_name, pnpm_scope
        #     )
        # else:
        yaml_lockfile, _, _ = get_lockfile(project_repo_name, release_version, package_manager)
        deps_list_all = extract_deps.extract_deps_from_pnpm_lockfile(project_repo_name, yaml_lockfile)

    # extract deps from lockfile
    elif package_manager == "yarn-classic" or package_manager == "yarn-berry":
        yarn_file, _, _ = get_lockfile(project_repo_name, release_version, package_manager)
        if package_manager == "yarn-classic":
            deps_list_all = extract_deps.extract_deps_from_v1_yarn(project_repo_name, yarn_file)
        elif package_manager == "yarn-berry":
            deps_list_all = deps_list_all = extract_deps.extract_deps_from_yarn_berry(project_repo_name, yarn_file)
            patches_info = extract_deps.get_patches_info(project_repo_name, yarn_file)

    elif package_manager == "npm":
        npm_file, _, _ = get_lockfile(project_repo_name, release_version, package_manager)
        deps_list_all = extract_deps.extract_deps_from_npm(project_repo_name, npm_file)

    elif package_manager == "maven":
        # Maven is more complex, because of child packages in the repo/pom; this requires to clone the whole repo
        # TODO: Issue: not sure if this works with child projects -- probably not?
        # Example: parent package A has a child package B; we want to run it on package B, but cloning won't work here (?)
        # And even if it did, we still need to, inside the project, navigate to the child package and run the analysis there
        # So this is a side case not yet handled
        repo_path = tool_config.clone_repo(project_repo_name, release_version)
        deps_list_all = extract_deps.extract_deps_from_maven(repo_path)

    logging.info("Number of dependencies: %d", len(deps_list_all.get("resolutions", {})))
    logging.info("Number of patches: %d", len(deps_list_all.get("patches", {})))
    logging.info("Number of aliased packages: %d", len(deps_list_all.get("aliased_packages", {})))
    logging.info("Number of workspace dependencies: %d", len(deps_list_all.get("workspace", {})))

    # dep with different resolutions for further analysis
    dep_with_many_versions = extract_deps.deps_versions(deps_list_all)
    logging.info(
        "Number of dependencies with different resolutions: %d",
        len(dep_with_many_versions),
    )

    return deps_list_all, dep_with_many_versions, patches_info


def static_analysis_all(
    folder_path,
    project_repo_name,
    release_version,
    package_manager,
    check_match=False,
    enabled_checks=None,
    config=None,
):
    """
    Perform static analysis on the given project and release version.

    Args:
        folder_path (str): The path to the project folder.
        project_repo_name (str): The name of the project repository.
        release_version (str): The release version of the project.
        package_manager (str): The package manager used in the project.
        check_match (bool): Whether to check for package name matches.
        enabled_checks (dict): Dictionary of enabled smell checks.
        config (dict): Configuration dictionary
    """
    deps_list, dep_with_many_versions, patches_info = get_deps(
        folder_path, project_repo_name, release_version, package_manager
    )
    repo_url_info = github_repo.get_github_repo_url(folder_path, deps_list, package_manager)

    static_results, errors = static_analysis.get_static_data(
        folder_path,
        repo_url_info,
        package_manager,
        check_match=check_match,
        enabled_checks=enabled_checks,
        config=config,
    )
    logging.info("Errors: %s", errors)

    return static_results, deps_list, dep_with_many_versions, patches_info


def differential_analysis(
    new_release_version,
    old_rv_dep_versions,
    new_rv_dep_versions,
    sa_1,
    sa_2,
    patches_new,
    patches_old,
    project_repo_name,
    package_manager,
):
    """
    Perform differential analysis on the given project and release versions.

    Args:
        new_release_version (str): The new release version of the project.
        old_rv_dep_versions (dict): The dependencies for the old release version.
        new_rv_dep_versions (dict): The dependencies for the new release version.
        sa_1 (dict): The static analysis results for the old release version.
        sa_2 (dict): The static analysis results for the new release version.
        patches_new (dict): The patches info for the new release version.
        patches_old (dict): The patches info for the old release version.
        project_repo_name (str): The name of the project repository.
        package_manager (str): The package manager used in the project.
    Returns:
        tuple: A tuple containing the following:
            - compare_differences (dict): The comparison results for the dependencies.
            - downgraded_pkg (dict): The downgraded packages.
            - upgraded_pkg (dict): The upgraded packages.
            - all_compare_info (dict): The comparison results for the PRs.
    """

    logging.info("Running differential analysis...")

    (
        compare_differences,
        differences_pkg_full,
        downgraded_pkg,
        _,
        _,
        upgraded_pkg,
        _,
    ) = compare_packages.differential(old_rv_dep_versions, new_rv_dep_versions, sa_1, sa_2)

    if package_manager != "maven":
        changed_patches, _ = compare_packages.changed_patch(patches_old, patches_new)
    else:
        changed_patches = {}

    authors = compare_commits.get_commit_results(
        headers,
        project_repo_name,
        new_release_version,
        changed_patches,
        differences_pkg_full,
    )

    commit_results = get_user_commit_info.get_user_first_commit_info(authors)

    useful_pr_infos = get_pr_info.get_useful_pr_info(commit_results)

    all_compare_info = get_pr_review.get_pr_review_info(useful_pr_infos)

    return (
        compare_differences,
        downgraded_pkg,
        upgraded_pkg,
        all_compare_info,
    )


def write_to_file(filename, directory, data):
    """
    Write data to a JSON file.

    Args:
        filename (str): The name of the file to write to.
        directory (str): The directory to write the file to.
        data (dict): The data to write to the file.
    """
    file_path = os.path.join(directory, filename)

    with open(file_path, "w", encoding="utf-8") as file:
        json.dump(data, file, indent=4)


def setup_project_info(args, any_check_specified):
    """Set up project information based on command-line arguments."""
    version_old, version_new = args.release_version_old, args.release_version_new
    if version_old in ["HEAD", "HEAD^", "main", "master"]:
        api_url = f"https://api.github.com/repos/{args.project_repo_name}/commits/{version_old}"
        data = tool_config.make_github_request(api_url)
        version_old = data.get("sha")
        if not version_old:
            logging.error(f"Failed to get the commit SHA for {version_old}.")
            raise ValueError(f"Failed to get the commit SHA for {version_old}.")

    if version_new and version_new in ["HEAD", "HEAD^", "main", "master"]:
        api_url = f"https://api.github.com/repos/{args.project_repo_name}/commits/{version_new}"
        data = tool_config.make_github_request(api_url)
        version_new = data.get("sha")
        if not version_new:
            logging.error(f"Failed to get the commit SHA for {version_new}.")
            raise ValueError(f"Failed to get the commit SHA for {version_new}.")

    return {
        "repo_name": args.project_repo_name,
        "old_version": version_old,
        "new_version": version_new,
        "old_version_name": version_old.replace("/", "_"),
        "new_version_name": (version_new.replace("/", "_") if version_new else None),
        "check_match": args.name_match,
        "package_manager": args.package_manager,
        # "pnpm_scope": args.pnpm_scope,
        "debug": args.debug,
        "gradual_report": args.gradual_report and not any_check_specified,
    }


def setup_directories_and_logging(project_info):
    """Set up necessary directories and logging."""

    dir_path = tool_config.PathManager()
    result_folder_path, json_directory, diff_folder = dir_path.create_folders(project_info["old_version_name"])
    project_info["result_folder_path"] = result_folder_path
    project_info["json_directory"] = json_directory
    project_info["diff_folder"] = diff_folder

    log_file_path = result_folder_path / "analysis.log"
    tool_config.setup_logger(log_file_path, project_info["debug"])


def verify_config(config):
    """Check whether there are conflicting ignored checks for packages in the configuration file."""
    for checks_per_type in (config.get("ignore", {}), config.get("ignore-if-parent", {})):
        for package, checks in checks_per_type.items():
            if isinstance(checks, str) and checks == "all":
                continue
            elif not isinstance(checks, list):
                raise ValueError(f"Invalid checks for package {package}: {checks}")

            disallowed_if_source_code_ignored = ["source_code_sha", "forks"]
            wrongly_unignored = [check for check in disallowed_if_source_code_ignored if check not in checks]
            if "source_code" in checks and any(wrongly_unignored):
                raise ValueError(
                    f"""Conflicting checks for package {package}: source_code is asked to be ignored, yet the following depend on it and are not: {wrongly_unignored}. Please remove source_code from the ignored checks or add the aforementioned checks to the ignored checks list."""
                )


def perform_static_analysis(project_info, is_old_version):
    """Perform static analysis for a given version."""

    version = project_info["old_version"] if is_old_version else project_info["new_version"]
    version_name = project_info["old_version_name"] if is_old_version else project_info["new_version_name"]
    json_dir = (
        project_info["json_directory"]
        if is_old_version
        else (project_info["result_folder_path"] / "details" / version_name)
    )

    if not is_old_version:
        json_dir.mkdir(parents=True, exist_ok=True)

    results = static_analysis_all(
        json_dir,
        project_info["repo_name"],
        version,
        project_info["package_manager"],
        project_info["check_match"],
        project_info["enabled_checks"],
        project_info["config"],
    )

    write_to_file(
        f"{version_name}_static_results.json",
        project_info["json_directory"],
        results[0],
    )

    return results


def generate_static_report(analysis_results, project_info, is_old_version):
    """Generate static analysis report."""
    version = project_info["old_version"] if is_old_version else project_info["new_version"]
    version_name = project_info["old_version_name"] if is_old_version else project_info["new_version_name"]

    summary_file = project_info["result_folder_path"] / f"{version_name}_static_summary.md"

    logging.info("Generating static analysis report for %s", version_name)
    report_static.get_s_summary(
        analysis_results[0],  # static analysis results
        analysis_results[1],  # deps_list
        project_info["repo_name"],
        version,
        project_info["package_manager"],
        project_info["enabled_checks"],
        project_info["gradual_report"],
        summary_filename=summary_file,
    )


def perform_differential_analysis(old_results, new_results, project_info):
    """Perform and report differential analysis."""
    diff_results = differential_analysis(
        project_info["new_version"],
        old_results[2],
        new_results[2],  # dep_with_many_versions
        old_results[0],
        new_results[0],  # static_results
        new_results[3],
        old_results[3],  # patches_info
        project_info["repo_name"],
        project_info["package_manager"],
    )

    # Write differential analysis results to files
    for filename, data in zip(
        [
            "compare_differences_pkg.json",
            "downgraded_pkg.json",
            "upgraded_pkg.json",
            "compare_prr_infos.json",
        ],
        diff_results,
    ):
        write_to_file(filename, project_info["diff_folder"], data)

    # Generate differential analysis report
    diff_summary = (
        project_info["result_folder_path"]
        / f"{project_info['old_version_name']}_{project_info['new_version_name']}_diff_summary.md"
    )
    logging.info("Generating differential analysis report...")
    report_diff.generate_diff_report(
        diff_results[3],  # diff_compare_all_info
        project_info["repo_name"],
        project_info["old_version_name"],
        project_info["new_version_name"],
        project_info["gradual_report"],
        diff_summary,
    )


def main():
    """Main flow to run the software supply chain smell analysis."""
    dw_args = get_args()

    # Determine which checks are enabled
    enabled_checks = {}
    any_check_specified = any(
        [
            dw_args.check_source_code,
            dw_args.check_source_code_sha,
            dw_args.check_deprecated,
            dw_args.check_forks,
            dw_args.check_provenance,
            dw_args.check_code_signature,
            dw_args.check_aliased_packages,
        ]
    )

    if any_check_specified:
        # Check if source code SHA or fork checks are enabled without source code
        if any([dw_args.check_source_code_sha, dw_args.check_forks]) and not dw_args.check_source_code:
            raise ValueError(
                "The --check-source-code-sha and --check-forks flags require --check-source-code to be enabled. "
                "These can only be checked if we can first verify the source code repository."
            )

        enabled_checks = {
            "source_code": dw_args.check_source_code,
            "source_code_sha": dw_args.check_source_code_sha,
            "deprecated": dw_args.check_deprecated,
            "forks": dw_args.check_forks,
            "provenance": dw_args.check_provenance,
            "code_signature": dw_args.check_code_signature,
            "aliased_packages": dw_args.check_aliased_packages,
        }
    else:
        # If no checks specified, enable all by default
        enabled_checks = {
            "source_code": True,
            "source_code_sha": True,
            "deprecated": True,
            "provenance": True,
            "code_signature": True,
            "aliased_packages": True,
        }

    project_info = setup_project_info(dw_args, any_check_specified)
    setup_directories_and_logging(project_info)
    project_info["config"] = tool_config.load_config(dw_args.config)
    verify_config(project_info["config"])

    logging.info(
        f"Software supply chain smells analysis for {project_info['repo_name']} for version {project_info['old_version']}..."
    )
    if project_info["gradual_report"]:
        logging.info("Gradual report generation is enabled.")

    # Pass enabled_checks to static analysis
    project_info["enabled_checks"] = enabled_checks

    old_analysis_results = perform_static_analysis(project_info, is_old_version=True)
    generate_static_report(old_analysis_results, project_info, is_old_version=True)

    if project_info["new_version"]:
        new_analysis_results = perform_static_analysis(project_info, is_old_version=False)
        generate_static_report(new_analysis_results, project_info, is_old_version=False)

        if dw_args.differential_analysis:
            perform_differential_analysis(old_analysis_results, new_analysis_results, project_info)


if __name__ == "__main__":
    main()
    logging.info("Analysis completed.")



"""Module for extracting dependencies.
Support npm, yarn classic, yarn berry, pnpm
"""

import re
import os
import subprocess
import json
import logging
import sys
import shutil
from collections import defaultdict
import json
import hashlib
from pathlib import Path
import yaml

from tool.tool_config import PNPM_LIST_COMMAND, get_cache_manager, YarnLockParser

cache_manager = get_cache_manager()

MVN_DEPENDENCY_PLUGIN = "org.apache.maven.plugins:maven-dependency-plugin:3.8.1"
append_dependency_goal = lambda goal: f"{MVN_DEPENDENCY_PLUGIN}:{goal}"
TREE_GOAL = append_dependency_goal("tree")
RESOLVE_PLUGINS_GOAL = append_dependency_goal("resolve-plugins")
TREE_LOG = "/tmp/deps.json"
RESOLVE_PLUGINS_LOG = "/tmp/plugins.log"


def get_lockfile_hash(lockfile_content):
    """Generate a hash of the lockfile to detect changes"""
    return hashlib.sha256(str(lockfile_content).encode()).hexdigest()


def extract_deps_from_pnpm_lockfile(repo_path, pnpm_lockfile_yaml):
    """
    Extract dependencies from a pnpm-lock.yaml file.

    Args:
        repo_path (str): The project's source code repository.
        pnpm_lockfile_yaml (str): The content of the pnpm lock file.

    Returns:
        dict: A dictionary containing the extracted dependencies and patches.
    """

    try:
        yaml_data = yaml.safe_load(pnpm_lockfile_yaml)
        yaml_version = yaml_data.get("lockfileVersion")
        if yaml_version != "9.0":
            logging.error("Invalid pnpm lockfile version: %s", yaml_version)
            logging.error("The pnpm lockfile version is not supported(yet): ", yaml_version)
            # end the process
            sys.exit(1)

        lockfile_hash = get_lockfile_hash(yaml_data)
        if not lockfile_hash:
            logging.error("No lockfile found in %s", repo_path)
            return {"resolutions": [], "patches": []}

        cached_deps = cache_manager.extracted_deps_cache.get_dependencies(repo_path, lockfile_hash)
        if cached_deps:
            logging.info(f"Using cached dependencies for {repo_path}")
            return cached_deps

        parent_packages = defaultdict(set)
        pkg_name_with_resolution = []
        patches = []

        # Iterate through packages to build parent-child relationships
        for pkg_name, pkg_info in yaml_data.get("snapshots", {}).items():
            version_match = re.search(r"@([^@]+)$", pkg_name)
            version = version_match.group(1) if version_match else "unknown"

            # Clean up package name (remove version)
            pkg_name = re.sub(r"@[^@]+$", "", pkg_name)

            # Construct resolution string
            resolution = f"{pkg_name}@{version}"
            pkg_name_with_resolution.append(resolution)

            # Track child dependencies
            if pkg_info.get("dependencies"):
                for child_name, child_version in pkg_info["dependencies"].items():
                    child_resolution = f"{child_name}@{child_version}"
                    parent_packages[child_resolution].add(resolution)

        # Convert to required format with parent information
        deps_list_data = {
            "resolutions": list(
                {
                    "info": info,
                    "parent": list(parent_packages.get(info, set())),
                }
                for info in sorted(pkg_name_with_resolution)
            ),
            "patches": patches,
        }

        cache_manager.extracted_deps_cache.cache_dependencies(repo_path, lockfile_hash, deps_list_data)

        return deps_list_data
    except yaml.YAMLError as e:
        logging.error(
            "An error occurred while parsing the pnpm-lock.yaml file: %s",
            str(e),
        )
        return {"resolutions": [], "patches": []}
    except (IOError, ValueError, KeyError) as e:
        logging.error(
            "An error occurred while extracting dependencies from pnpm-lock.yaml: %s",
            str(e),
        )
        return {"resolutions": [], "patches": []}


def extract_deps_from_npm(repo_path, npm_lock_file):
    """
    Extract dependencies from a "package-lock.json" file.

    Args:
        repo_path (str): The project's source code repository.
        npm_lock_file (dict): The content of the npm lock file.

    Returns:
        dict: A dictionary containing the extracted dependencies and patches.
    """
    lock_file_json = json.loads(npm_lock_file)
    lockfile_hash = get_lockfile_hash(lock_file_json)
    if not lockfile_hash:
        logging.error("No lockfile found in %s", repo_path)
        return {"resolutions": [], "patches": []}

    cached_deps = cache_manager.extracted_deps_cache.get_dependencies(repo_path, lockfile_hash)
    if cached_deps:
        logging.info(f"Using cached dependencies for {repo_path}")
        return cached_deps
    try:
        patches = []
        pkg_name_with_resolution = set()
        aliased_packages = {}
        deps_list_data = {}

        parent_packages = {}
        if lock_file_json.get("packages") and isinstance(lock_file_json["packages"], dict):
            for package_path, package_info in lock_file_json["packages"].items():
                if package_path.startswith("node_modules/"):
                    package_name = package_path.split("/", 1)[1]
                    if "node_modules" in package_name:
                        package_name = package_name.split("node_modules/")[-1]

                    resolution = package_name
                    if package_info.get("version"):
                        version = package_info["version"]
                        # Handle npm aliases
                        original_name = package_info.get("name")
                        if original_name:
                            logging.warning(f"Found npm alias for {original_name}@{version}")
                            aliased_packages[f"{original_name}@{version}"] = package_name
                            package_name = original_name

                        resolution = f"{package_name}@{version}"
                        pkg_name_with_resolution.add(resolution)

                    if package_info.get("dependencies"):
                        for dep_name, version in package_info["dependencies"].items():
                            parent_packages.setdefault(f"{dep_name}@{version}", set()).add(resolution)

            deps_list_data = {
                "resolutions": list(
                    {
                        "info": info,
                        "parent": list(parent_packages.get(info, set())),
                    }
                    for info in sorted(pkg_name_with_resolution)
                ),
                "patches": patches,
                "aliased_packages": aliased_packages,
            }

            cache_manager.extracted_deps_cache.cache_dependencies(repo_path, lockfile_hash, deps_list_data)

        return deps_list_data

    except (IOError, ValueError, KeyError) as e:
        logging.error(
            "An error occurred while extracting dependencies from package-lock.json: %s",
            str(e),
        )

    return {"resolutions": [], "patches": [], "aliased_packages": []}


def extract_deps_from_yarn_berry(repo_path, yarn_lock_file):
    """
    # JavaScript
    Extract dependencies from a Yarn Berry lock file.

    Args:
        repo_path (str): The project's source code repository.
        yarn_lock_file (str): The content of the Yarn Berry lock file.

    Returns:
        dict: A dictionary containing the extracted dependencies and patches.
    """
    lockfile_hash = get_lockfile_hash(yarn_lock_file)
    if not lockfile_hash:
        logging.error("No lockfile found in %s", repo_path)
        return {"resolutions": [], "patches": []}

    cached_deps = cache_manager.extracted_deps_cache.get_dependencies(repo_path, lockfile_hash)
    if cached_deps:
        logging.info(f"Using cached dependencies for {repo_path}")
        return cached_deps

    try:
        patches = []
        pkg_name_with_resolution = []
        aliased_packages = {}

        parent_packages = {}
        parsed_lockfile = yaml.safe_load(yarn_lock_file)
        for entry_data in parsed_lockfile.values():
            resolution = entry_data.get("resolution", "")
            if resolution:
                if "@patch" in resolution:
                    # Check if "patch" is present in the resolution
                    resolution = resolution.replace('resolution: "', "").strip('"').lstrip()
                    patches.append(resolution)
                else:
                    # aliases will show up as something like my-foo@npm:foo@x.y.z
                    alias_pattern = r"(.+?)@npm:(.+?)@(.+)"
                    alias_match = re.match(alias_pattern, resolution)
                    if alias_match:
                        # if it is an alias, we add the original name to the list
                        logging.info(f"Found yarn alias for {alias_match.group(2)}@{alias_match.group(3)}")
                        logging.info(f"Aliased to {alias_match.group(1)}@{alias_match.group(3)}")
                        aliased_packages[f"{alias_match.group(2)}@{alias_match.group(3)}"] = (
                            f"{alias_match.group(1)}@{alias_match.group(3)}"
                        )
                        resolution = f"{alias_match.group(2)}@{alias_match.group(3)}"
                    pkg_name_with_resolution.append(resolution)

                if entry_data.get("dependencies"):
                    for dep_name, version in entry_data["dependencies"].items():
                        parent_packages.setdefault(f"{dep_name}@{version}", set()).add(resolution)

        deps_list_data = {
            "resolutions": list(
                {"info": info, "parent": list(parent_packages.get(info, set()))}
                for info in sorted(pkg_name_with_resolution)
            ),
            "patches": list({"info": info} for info in sorted(patches)),
            "aliased_packages": aliased_packages,
        }

        cache_manager.extracted_deps_cache.cache_dependencies(repo_path, lockfile_hash, deps_list_data)

        return deps_list_data

    except yaml.YAMLError as e:
        logging.error(
            "An error occurred while parsing the yarn.lock file(Yarn Berry): %s",
            str(e),
        )
        return {"resolutions": [], "patches": [], "aliased_packages": []}
    except (IOError, ValueError, KeyError) as e:
        logging.error(
            "An error occurred while extracting dependencies from yarn.lock file(Yarn Berry): %s",
            str(e),
        )
        return {"resolutions": [], "patches": [], "aliased_packages": []}


def extract_deps_from_v1_yarn(repo_path, yarn_lock_file):
    """
    # JavaScript
    Extract dependencies from a Yarn Classic lock file.

    Args:
        repo_path (str): The project's source code repository.
        yarn_lock_file (str): The content of the Yarn Classic lock file.

    Returns:
        dict: A dictionary containing the extracted dependencies and patches.
    """
    # yarn-classic
    lockfile_hash = get_lockfile_hash(yarn_lock_file)
    if not lockfile_hash:
        logging.error("No lockfile found in %s", repo_path)
        return {"resolutions": [], "patches": []}

    cached_deps = cache_manager.extracted_deps_cache.get_dependencies(repo_path, lockfile_hash)
    if cached_deps:
        logging.info(f"Using cached dependencies for {repo_path}")
        return cached_deps
    try:
        pkg_name_with_resolution = []
        patches = []
        aliased_packages = {}
        parent_packages = defaultdict(set)

        # Find all dependencies
        parser = YarnLockParser(yarn_lock_file)
        dependencies = parser.parse()
        for entry_name, entry_data in dependencies.items():
            item = f"{entry_name}@{entry_data['version_constraint']}"
            if entry_data.get("original_name"):
                logging.warning(f"Found yarn alias for {item}")
                logging.warning(f"Original name: {entry_data['original_name']}@{entry_data['version_constraint']}")
                logging.warning(f"Aliased to: {entry_name}@{entry_data['version_constraint']}")
                aliased_packages[item] = f"{entry_data['original_name']}@{entry_data['version_constraint']}"

            if entry_data.get("dependencies"):
                logging.info("Found child dependencies for %s", item)
                for dep_name, dep_version in entry_data["dependencies"].items():
                    dep_name = f"{dep_name}@{dep_version}"
                    parent_packages[dep_name].add(item)

            pkg_name_with_resolution.append(item)

        deps_list_data = {
            "resolutions": list(
                {
                    "info": info,
                    "parent": list(parent_packages.get(info, set())),
                }
                for info in sorted(pkg_name_with_resolution)
            ),
            "patches": patches,
            "aliased_packages": aliased_packages,
        }

        cache_manager.extracted_deps_cache.cache_dependencies(repo_path, lockfile_hash, deps_list_data)

        return deps_list_data

    except (IOError, ValueError, KeyError) as e:
        logging.error(
            "An error occurred while extracting dependencies from yarn.lock file(Yarn Classic): %s",
            str(e),
        )
        return {"resolutions": [], "patches": [], "aliased_packages": []}


def get_pnpm_dep_tree(folder_path, version_tag, repo_path, pnpm_scope):
    """
    Get pnpm dependency tree for the given project.

    Args:
        folder_path (str): Path to the project folder
        version_tag (str): Version tag of the project
        repo_path (str): Name of the project repository
        pnpm_scope (str): Scope of the pnpm package

    Returns:
        dict: Dependency tree
    """
    version_tag_name = version_tag.replace("/", "-")
    repo_path_for_dir = repo_path.replace("/", "-")
    # for pnpm mono repo
    dir_name = f"{repo_path_for_dir}-{version_tag_name}"

    if not os.path.exists(dir_name):
        repo_url = f"https://github.com/{repo_path}.git"
        subprocess.run(
            [
                "git",
                "clone",
                "--depth",
                "1",
                "--branch",
                f"{version_tag}",
                repo_url,
                dir_name,
            ],
            check=True,
        )
        logging.info("Cloning %s Repository...", repo_path)

        # repo_name = repo_url.split("/")[-1].replace(".git", "")

        logging.info("Cloned to %s", dir_name)

        os.chdir(dir_name)

        logging.info("Installing dependencies...")
        # try:
        subprocess.run(["pnpm", "install", "--ignore-scripts"], check=True)

    try:
        # if it is not in dir_name, change to dir_name
        dir_if = os.getcwd().split("/")[-1]
        if dir_if != dir_name:
            os.chdir(dir_name)

        logging.info("Getting pnpm dependency tree by running %s", PNPM_LIST_COMMAND(pnpm_scope))

        command = PNPM_LIST_COMMAND(pnpm_scope)
        logging.info("Getting pnpm dependency tree...")
        result = subprocess.run(
            command,
            check=True,
            capture_output=True,
            text=True,
        )

        parent_directory = os.path.abspath("..")
        details_folder = os.path.join(parent_directory, folder_path)

        output_file_path = os.path.join(details_folder, "pnpm_dep_tree.txt")
        logging.info("Writing pnpm dependency tree to %s", output_file_path)

        with open(output_file_path, "w", encoding="utf-8") as f:
            f.write(result.stdout)

        # os.chdir("..")
        change_to_cloned_parent_dir = f"../{dir_name}"
        shutil.rmtree(change_to_cloned_parent_dir)
        logging.info("Removed %s", dir_name)
        # change to child folder

        return result.stdout.splitlines(), folder_path

    except subprocess.CalledProcessError as e:
        logging.error(f"An error occurred: {e}")
        sys.exit(1)

    finally:
        if os.path.exists(dir_name):
            shutil.rmtree(dir_name)
            logging.info("Removed %s", dir_name)


def extract_deps_from_pnpm_mono(folder_path, version_tag, repo_path, pnpm_scope):
    """
    Extract dependencies from a pnpm monorepo.

    Args:
        folder_path (str): Path to the monorepo folder
        version_tag (str): Version tag to use
        repo_path (str): Name of the project repository
        pnpm_scope (str): Scope of the pnpm package

    Returns:
        tuple: Lists of all dependencies and registry dependencies
    """
    all_deps = []
    registry_dep = []
    workspace_dep = []
    patched_dep = []

    tree, folder_path_for_this = get_pnpm_dep_tree(folder_path, version_tag, repo_path, pnpm_scope)
    lockfile_hash = get_lockfile_hash(tree)  # not really a lockfile, but an approximation
    os.chdir("..")
    if not lockfile_hash:
        logging.error("No lockfile found in %s", repo_path)
        return {"resolutions": [], "patches": []}

    cached_deps = cache_manager.extracted_deps_cache.get_dependencies(repo_path, lockfile_hash)
    if cached_deps:
        logging.info(f"Using cached dependencies for {repo_path}")
        return cached_deps

    os.chdir("..")
    if tree is None:
        return {}
    dependencies = defaultdict(list)

    dep_pattern = re.compile(r"([a-zA-Z0-9@._/-]+)\s+(\S+)")
    logging.info("Extracting dependencies from pnpm list output")

    for line in tree:
        # if "ledger-live-desktop" in line or "production dependency, optional only, dev only" in line:
        if pnpm_scope in line or "production dependency, optional only, dev only" in line:
            logging.info("ledger-live-desktop found")
            continue

        match = dep_pattern.search(line)
        if match:
            dep_name = match.group(1).strip()
            dep_version = match.group(2).strip()
            dependencies[dep_name].append(dep_version)
            logging.info("dependency found: %s", dep_name)

        # logging.info(f"Number of dependencies({version_tag}): {len(dependencies)}")

    for dep, versions in sorted(dependencies.items()):
        for version in sorted(set(versions)):
            formatted_dep = f"{dep}@{version}"
            all_deps.append(formatted_dep)
            if "link:" in formatted_dep:
                workspace_dep.append(formatted_dep)
            elif "patch_hash=" in formatted_dep:
                patched_dep.append(formatted_dep)
            else:
                registry_dep.append(formatted_dep)

    dep_all_path = os.path.join(folder_path_for_this, "deps_list_all_installed.json")

    logging.info("Writing all dependencies to %s", dep_all_path)

    with open(dep_all_path, "w", encoding="utf-8") as f:
        json.dump(all_deps, f, indent=4)

    deps_list_data = {
        "resolutions": list({"info": info} for info in registry_dep),
        "patches": list({"info": info} for info in patched_dep),
        "workspace": workspace_dep,
    }

    cache_manager.extracted_deps_cache.cache_dependencies(repo_path, lockfile_hash, deps_list_data)

    logging.info(
        "Number of packages from registry(%s): %d",
        version_tag,
        len(deps_list_data["resolutions"]),
    )
    logging.info(
        "Number of packages from workspace(%s): %d",
        version_tag,
        len(deps_list_data["workspace"]),
    )
    logging.info(
        "Number of patched packages(%s): %d",
        version_tag,
        len(deps_list_data["patches"]),
    )

    return deps_list_data


def get_pom_hash(repo_path):
    """Generate a hash of the pom.xml file to detect changes"""
    pom_path = Path(repo_path) / "pom.xml"
    if not pom_path.exists():
        return None

    with open(pom_path, "rb") as f:
        return hashlib.sha256(f.read()).hexdigest()


def extract_deps_from_maven(repo_path):
    """
    Extract dependencies from a Maven package, given the path to its locally cloned repo.

    Args:
        repo_path (str): The path to the locally cloned Maven package.

    Returns:
        dict: A dictionary containing the extracted dependencies.
    """

    def parse_mvn_tree_logs(log_file):
        """
        Parse Maven dependency tree logs to extract dependency information.

        Args:
            log_file (str): Path to the Maven dependency tree log file

        Returns:
            list: List of dictionaries containing dependency information
        """
        dependencies = []

        def parse_dependency(dep, parent=None):
            dep_info = {
                "groupId": dep["groupId"],
                "artifactId": dep["artifactId"],
                "version": dep["version"],
                "parent": f"{parent['groupId']}:{parent['artifactId']}@{parent['version']}" if parent else None,
            }
            dependencies.append(dep_info)
            for child in dep.get("children", []):
                parse_dependency(child, dep_info)

        try:
            with open(log_file, "r") as f:
                data = json.load(f)
                for dep in data["children"]:
                    parse_dependency(dep)

        except FileNotFoundError:
            logging.error("Dependency log file not found: %s", log_file)
        except Exception as e:
            logging.error("Error parsing dependency log: %s", str(e))

        return dependencies

    def parse_mvn_plugin_logs(log_file):
        """
        Parse Maven dependency resolution logs to extract dependency information.

        Args:
            log_file (str): Path to the Maven dependency resolution log file

        Returns:
            list: List of dictionaries containing dependency information
        """
        dependencies = []
        current_parent = None
        base_indent_level = 3  # 3 spaces is the base indent level
        try:
            with open(log_file, "r") as f:
                lines = f.readlines()
                parsing = False

                for line in lines:
                    if "The following plugins have been resolved:" in line:
                        parsing = True
                        continue

                    if "The following plugins have been" in line and parsing:
                        break

                    if parsing:
                        indent_level = len(line) - len(line.lstrip(" ")) - base_indent_level
                        parts = line.strip().split(":")
                        if len(parts) >= 3:
                            dep_info = {
                                "groupId": parts[0],
                                "artifactId": parts[1],
                                "version": parts[-1].split()[0],
                                "parent": None,
                            }
                            if indent_level == 0:
                                current_parent = (
                                    f"{dep_info['groupId']}:{dep_info['artifactId']}@{dep_info['version']}"
                                )
                            else:
                                dep_info["parent"] = current_parent
                                dependencies.append(dep_info)

        except FileNotFoundError:
            logging.error("Dependency log file not found: %s", log_file)
        except Exception as e:
            logging.error("Error parsing dependency log: %s", str(e))

        return dependencies

    # Generate a cache key based on the repo path and pom.xml hash
    pom_hash = get_pom_hash(repo_path)
    if not pom_hash:
        logging.error("No pom.xml found in %s", repo_path)
        return {"resolutions": [], "patches": []}

    cached_deps = cache_manager.extracted_deps_cache.get_dependencies(repo_path, pom_hash)
    if cached_deps:
        logging.info(f"Using cached Maven dependencies for {repo_path}")
        return cached_deps

    # If we reach here, we need to resolve dependencies
    current_dir = os.getcwd()
    os.chdir(repo_path)

    retrieval_commands = {
        "regular": [
            "mvn",
            TREE_GOAL,
            "-DoutputType=json",
            f"-DoutputFile={TREE_LOG}",
        ],
        "plugins": [
            "mvn",
            RESOLVE_PLUGINS_GOAL,
            "-Dsort=true",
            f"-DoutputFile={RESOLVE_PLUGINS_LOG}",
        ],
    }

    try:
        # Run Maven commands to resolve dependencies
        subprocess.run(retrieval_commands["regular"], check=True)
        subprocess.run(retrieval_commands["plugins"], check=True)

        # Parse the dependency logs
        retrieved_deps = parse_mvn_tree_logs(TREE_LOG)
        retrieved_plugins = parse_mvn_plugin_logs(RESOLVE_PLUGINS_LOG)

        # Go back to original directory
        os.chdir(current_dir)

        # Format the dependencies
        parsed_deps = [
            {
                "info": f"{dep['groupId']}:{dep['artifactId']}@{dep['version']}",
                "parent": dep["parent"],
                "command": "tree",
            }
            for dep in retrieved_deps
        ]
        parsed_plugins = [
            {
                "info": f"{plugin['groupId']}:{plugin['artifactId']}@{plugin['version']}",
                "parent": plugin["parent"],
                "command": "resolve-plugins",
            }
            for plugin in retrieved_plugins
        ]

        # Create the result
        deps_list_data = {
            "resolutions": list({item["info"]: item for item in parsed_plugins + parsed_deps}.values()),
            "patches": [],
        }

        # Cache the results
        cache_manager.extracted_deps_cache.cache_dependencies(repo_path, pom_hash, deps_list_data)

        return deps_list_data

    except subprocess.CalledProcessError as e:
        os.chdir(current_dir)
        logging.error("Error resolving Maven dependencies: %s", str(e))
        return {"resolutions": [], "patches": []}


def deps_versions(deps_versions_info_list):
    """
    Extract dependencies versions from a list of dependency versions.

    Args:
        deps_versions_info_list (list): List of dependency versions

    Returns:
        dict: Dictionary containing dependency names and their versions
    """
    deps_versions_dict = {}
    for pkg in deps_versions_info_list.get("resolutions"):
        pkg_name, version = pkg["info"].rsplit("@", 1)

        version = version.replace("npm:", "")

        if pkg_name in deps_versions_dict:
            deps_versions_dict[pkg_name].append(version)
        else:
            deps_versions_dict[pkg_name] = [version]

    return deps_versions_dict


def get_patches_info(repo_path, yarn_lock_file):
    """
    Get patches information from a Yarn Berry lock file.

    Args:
        repo_path (str): The project's source code repository.
        yarn_lock_file (str): The content of the Yarn Berry lock file.

    Returns:
        dict: Dictionary containing patch information
    """
    deps_list = extract_deps_from_yarn_berry(repo_path, yarn_lock_file)

    patches_info = {}
    for pkg_patches in deps_list.get("patches"):
        pattern = r"\.yarn/patches/(.*?\.patch).*version=([0-9.]+)&hash=([a-z0-9]+)"
        match = re.search(pattern, pkg_patches["info"])

        if match:
            patch_file_path = match.group(1)
            version = match.group(2)
            hash_code = match.group(3)

            patches_info[pkg_patches["info"]] = {
                "version": str(version),
                "hash_code": str(hash_code),
                "patch_file_path": str(patch_file_path),
            }

        else:
            patches_info[pkg_patches["info"]] = {
                "version": None,
                "hash_code": None,
                "patch_file_path": None,
            }

    logging.info("Number of patches: %d", len(patches_info))

    return patches_info
